
function butterfly_gen(p)


fid = fopen('../verilog/FFT_BUTTERFLY.v','w');

% -- wb_[
fprintf(fid,'`ifdef SYNTOOL_OFF \n');
fprintf(fid,'`else \n');
fprintf(fid,'`include "definition.v" \n');
fprintf(fid,'`endif \n\n');

for s=1:log2(p.N)
    fprintf(fid,'\n');
    fprintf(fid,'module BUTTERFLY_STAGE%d(CLK, k, X_r, X_i, Y_r, Y_i);\n', s);
    
    fprintf(fid,'input  CLK;\n');
    fprintf(fid,'input  [`N_AD_BUS-1:0]  k;\n');
    fprintf(fid,'input  [`W-1:0]  X_r, X_i;\n');
    fprintf(fid,'output [`W-1:0]  Y_r, Y_i;\n');
    fprintf(fid,'reg    [`W-1:0]  Y_r, Y_i;\n');
    fprintf(fid,'wire   [`W-1:0]  A_r, A_i, B_r, B_i;\n');
    fprintf(fid,'wire   [`W-1:0]  Fin_r, Fin_i;\n');
    fprintf(fid,'wire   [`W-1:0]  Fout_r, Fout_i;\n');
    fprintf(fid,'wire   [`W-1:0]  Min_r,  Min_i;\n');
    
    if (s < log2(p.N))
        fprintf(fid,'wire   [`W-1:0]  Mbf_r, Mbf_i;\n');
        fprintf(fid,'wire   [`W-1:0]  Mout_r, Mout_i;\n');   
        fprintf(fid,'wire   [`W-1:0]  W_r, W_i; \n');    
        fprintf(fid,'wire   [1:0]     W_ex_r, W_ex_i;\n');
        fprintf(fid,'wire			 scaling_bit = 1;\n');
    else
        fprintf(fid,'wire   [`W-1:0]  Mbf_r, Mbf_i;\n');
        fprintf(fid,'wire			 scaling_bit = 0;\n');
    end
    
    fprintf(fid,'assign Min_r = (k[%d] == 0) ? Fout_r : A_r;\n',log2(p.N)-s);
    fprintf(fid,'assign Min_i = (k[%d] == 0) ? Fout_i : A_i;\n',log2(p.N)-s);   
    fprintf(fid,'assign Fin_r = (k[%d] == 0) ? X_r : B_r;\n',log2(p.N)-s);
    fprintf(fid,'assign Fin_i = (k[%d] == 0) ? X_i : B_i;\n',log2(p.N)-s);
    
    if (s < log2(p.N))
        fprintf(fid,'assign Mbf_r = (k[%d] == 1) ? A_r : Mout_r;\n',log2(p.N)-s);   
        fprintf(fid,'assign Mbf_i = (k[%d] == 1) ? A_i : Mout_i;\n',log2(p.N)-s);
        fprintf(fid,'FIFO_ST%d FIFO_ST%d(CLK, Fin_r, Fin_i, Fout_r, Fout_i);\n',s,s);
        if (log2(p.N)-s-1==0)
            fprintf(fid,'TWIDDLE_ST%d TWIDDLE_ST%d(k[%d], W_ex_r, W_r, W_ex_i, W_i);\n',s,s,log2(p.N)-s-1);   
        else
            fprintf(fid,'TWIDDLE_ST%d TWIDDLE_ST%d(k[%d:0], W_ex_r, W_r, W_ex_i, W_i);\n',s,s,log2(p.N)-s-1);  
        end
        fprintf(fid,'CADD CADD(Fout_r, Fout_i, X_r, X_i, A_r, A_i, B_r, B_i, scaling_bit);\n');    
        fprintf(fid,'CMUL CMUL(Min_r, Min_i, W_r, W_ex_r, W_i, W_ex_i, Mout_r, Mout_i);\n');
    else
        fprintf(fid,'assign Mbf_r = Min_r;\n');   
        fprintf(fid,'assign Mbf_i = Min_i;\n');       
        fprintf(fid,'FIFO_ST%d FIFO_ST%d(CLK, Fin_r, Fin_i, Fout_r, Fout_i);\n',s,s);
        fprintf(fid,'CADD CADD(Fout_r, Fout_i, X_r, X_i, A_r, A_i, B_r, B_i, scaling_bit);\n');
    end    
    
    fprintf(fid,'always@(posedge CLK)\n');
    fprintf(fid,'begin\n');   
    fprintf(fid,'    Y_r <= Mbf_r;\n');
    fprintf(fid,'    Y_i <= Mbf_i;\n');
    fprintf(fid,'end\n');   
    % -- tb_[
    fprintf(fid,'endmodule\n\n');
end

fprintf(fid,'module CADD(X_r, X_i, Y_r, Y_i, A_r, A_i, B_r, B_i, scaling_bit);\n');
fprintf(fid,'input  signed [`W-1:0]  X_r, X_i, Y_r, Y_i;\n');
fprintf(fid,'output signed [`W-1:0]  A_r, A_i, B_r, B_i;\n');
fprintf(fid,'input  scaling_bit;\n');
fprintf(fid,'wire   signed [`W:0]  m_r, m_i, n_r, n_i;\n');
fprintf(fid,'assign m_r = X_r + Y_r;\n');
fprintf(fid,'assign m_i = X_i + Y_i;\n');
fprintf(fid,'assign n_r = X_r - Y_r;\n');
fprintf(fid,'assign n_i = X_i - Y_i;\n');
fprintf(fid,'assign A_r = (scaling_bit == 1) ? m_r >>> 1 : m_r;\n');
fprintf(fid,'assign A_i = (scaling_bit == 1) ? m_i >>> 1 : m_i;\n');
fprintf(fid,'assign B_r = (scaling_bit == 1) ? n_r >>> 1 : n_r;\n');
fprintf(fid,'assign B_i = (scaling_bit == 1) ? n_i >>> 1 : n_i;\n');
fprintf(fid,'endmodule\n');

fprintf(fid,'module CMUL(X_r, X_i, W_r, W_ex_r, W_i, W_ex_i, Y_r, Y_i);\n');
fprintf(fid,'input  [`W-1:0]  X_r, X_i;\n');
fprintf(fid,'input  [`W-1:0]  W_r, W_i;\n');
fprintf(fid,'input  [1:0]   W_ex_r, W_ex_i;\n');
fprintf(fid,'output [`W-1:0]  Y_r, Y_i;\n');
fprintf(fid,'wire   [2*`W-2:0]  m_rr, m_ri, m_ir, m_ii;\n');
fprintf(fid,'wire   signed [2*`W-2:0]  n_rr, n_ri, n_ir, n_ii;\n');
fprintf(fid,'wire   [2*`W-1:0]  o_r, o_i;\n');
fprintf(fid,'wire   signed [2*`W-2:0]  a_r, a_i;\n');
fprintf(fid,'wire 	zero = 0;\n');
fprintf(fid,'assign m_rr = FIX_MUL(W_r, X_r);\n');
fprintf(fid,'assign m_ri = FIX_MUL(W_i, X_r);\n');
fprintf(fid,'assign m_ir = FIX_MUL(W_r, X_i);\n');
fprintf(fid,'assign m_ii = FIX_MUL(W_i, X_i);\n');
fprintf(fid,'assign a_r  = {X_r, {`W-1{zero}}};\n');
fprintf(fid,'assign a_i  = {X_i, {`W-1{zero}}};\n');
fprintf(fid,'assign n_rr = (W_ex_r == 1) ?   a_r :\n');
fprintf(fid,'              (W_ex_r == 2) ?  -a_r :\n');
fprintf(fid,'                                m_rr;\n');
fprintf(fid,'assign n_ri = (W_ex_i == 1) ?   a_r :\n');
fprintf(fid,'              (W_ex_i == 2) ?  -a_r :\n');
fprintf(fid,'                                m_ri;\n');
fprintf(fid,'assign n_ir = (W_ex_r == 1) ?   a_i :\n');
fprintf(fid,'              (W_ex_r == 2) ?  -a_i :\n');
fprintf(fid,'                                m_ir;\n');
fprintf(fid,'assign n_ii = (W_ex_i == 1) ?   a_i :\n');
fprintf(fid,'              (W_ex_i == 2) ?  -a_i :\n');
fprintf(fid,'                                m_ii;\n');
fprintf(fid,'assign o_r = n_rr - n_ii;\n');
fprintf(fid,'assign o_i = n_ir + n_ri;\n');
fprintf(fid,'assign Y_r = o_r[2*`W-2:`W-1];\n');
fprintf(fid,'assign Y_i = o_i[2*`W-2:`W-1];\n');
fprintf(fid,'function [2*`W-2:0] FIX_MUL;\n');
fprintf(fid,'	input signed [`W-1:0] in1;\n');
fprintf(fid,'	input signed [`W-1:0] in2;\n');
fprintf(fid,'	reg signed [2*`W-2:0] qs;\n');	
fprintf(fid,'	begin\n');
fprintf(fid,'   	qs  = in1 * in2;\n');
fprintf(fid,'   	FIX_MUL = qs;\n');
fprintf(fid,'end\n');	
fprintf(fid,'endfunction\n');
fprintf(fid,'endmodule\n');

fclose(fid);