`ifdef SYNTOOL_OFF 
`else 
`include "definition.v" 
`endif 

module FFT(CLK, RESET, EN, X_r, X_i, Y_r, Y_i);
input          CLK;
input          RESET;
input          EN;
parameter      IDLE = 0;
parameter      RUN  = 1;
input  [`W-1:0]  X_r, X_i;
output [`W-1:0]  Y_r, Y_i;
reg              state;
reg    [`N_AD_BUS-1:0]   k_count_st1;
reg    [`N_AD_BUS-1:0]   k_count_st2;
reg    [`N_AD_BUS-1:0]   k_count_st3;
reg    [`N_AD_BUS-1:0]   k_count_st4;
reg    [`N_AD_BUS-1:0]   k_count_st5;
reg    [`N_AD_BUS-1:0]   k_count_st6;
wire   [`W-1:0]  Y_st1_r, Y_st1_i;
wire   [`W-1:0]  Y_st2_r, Y_st2_i;
wire   [`W-1:0]  Y_st3_r, Y_st3_i;
wire   [`W-1:0]  Y_st4_r, Y_st4_i;
wire   [`W-1:0]  Y_st5_r, Y_st5_i;
wire   [`W-1:0]  Y_st6_r, Y_st6_i;
BUTTERFLY_STAGE1 BUTTERFLY_STAGE1(CLK, k_count_st1,      X_r,     X_i, Y_st1_r, Y_st1_i);
BUTTERFLY_STAGE2 BUTTERFLY_STAGE2(CLK, k_count_st2,  Y_st1_r, Y_st1_i, Y_st2_r, Y_st2_i);
BUTTERFLY_STAGE3 BUTTERFLY_STAGE3(CLK, k_count_st3,  Y_st2_r, Y_st2_i, Y_st3_r, Y_st3_i);
BUTTERFLY_STAGE4 BUTTERFLY_STAGE4(CLK, k_count_st4,  Y_st3_r, Y_st3_i, Y_st4_r, Y_st4_i);
BUTTERFLY_STAGE5 BUTTERFLY_STAGE5(CLK, k_count_st5,  Y_st4_r, Y_st4_i, Y_st5_r, Y_st5_i);
BUTTERFLY_STAGE6 BUTTERFLY_STAGE6(CLK, k_count_st6,  Y_st5_r, Y_st5_i, Y_r,     Y_i);
always@(posedge CLK or negedge RESET)
begin
 if(!RESET)
    begin
     state   <= IDLE;
    end
 else
   begin
     k_count_st2 <= k_count_st1;
     k_count_st3 <= k_count_st2;
     k_count_st4 <= k_count_st3;
     k_count_st5 <= k_count_st4;
     k_count_st6 <= k_count_st5;
     case(state)
        IDLE :
           begin
             k_count_st1 <= 0;
             if(EN == 1) state <= RUN;
             else           state <= IDLE;
           end
        RUN :
           begin
             k_count_st1 <= k_count_st1 + 1;
             if(EN == 0)  state <= IDLE;
           end
      endcase
   end
end
endmodule
