`ifdef SYNTOOL_OFF 
`else 
`include "definition.v" 
`endif 


module BUTTERFLY_STAGE1(CLK, k, X_r, X_i, Y_r, Y_i);
input  CLK;
input  [`N_AD_BUS-1:0]  k;
input  [`W-1:0]  X_r, X_i;
output [`W-1:0]  Y_r, Y_i;
reg    [`W-1:0]  Y_r, Y_i;
wire   [`W-1:0]  A_r, A_i, B_r, B_i;
wire   [`W-1:0]  Fin_r, Fin_i;
wire   [`W-1:0]  Fout_r, Fout_i;
wire   [`W-1:0]  Min_r,  Min_i;
wire   [`W-1:0]  Mbf_r, Mbf_i;
wire   [`W-1:0]  Mout_r, Mout_i;
wire   [`W-1:0]  W_r, W_i; 
wire   [1:0]     W_ex_r, W_ex_i;
wire			 scaling_bit = 1;
assign Min_r = (k[5] == 0) ? Fout_r : A_r;
assign Min_i = (k[5] == 0) ? Fout_i : A_i;
assign Fin_r = (k[5] == 0) ? X_r : B_r;
assign Fin_i = (k[5] == 0) ? X_i : B_i;
assign Mbf_r = (k[5] == 1) ? A_r : Mout_r;
assign Mbf_i = (k[5] == 1) ? A_i : Mout_i;
FIFO_ST1 FIFO_ST1(CLK, Fin_r, Fin_i, Fout_r, Fout_i);
TWIDDLE_ST1 TWIDDLE_ST1(k[4:0], W_ex_r, W_r, W_ex_i, W_i);
CADD CADD(Fout_r, Fout_i, X_r, X_i, A_r, A_i, B_r, B_i, scaling_bit);
CMUL CMUL(Min_r, Min_i, W_r, W_ex_r, W_i, W_ex_i, Mout_r, Mout_i);
always@(posedge CLK)
begin
    Y_r <= Mbf_r;
    Y_i <= Mbf_i;
end
endmodule


module BUTTERFLY_STAGE2(CLK, k, X_r, X_i, Y_r, Y_i);
input  CLK;
input  [`N_AD_BUS-1:0]  k;
input  [`W-1:0]  X_r, X_i;
output [`W-1:0]  Y_r, Y_i;
reg    [`W-1:0]  Y_r, Y_i;
wire   [`W-1:0]  A_r, A_i, B_r, B_i;
wire   [`W-1:0]  Fin_r, Fin_i;
wire   [`W-1:0]  Fout_r, Fout_i;
wire   [`W-1:0]  Min_r,  Min_i;
wire   [`W-1:0]  Mbf_r, Mbf_i;
wire   [`W-1:0]  Mout_r, Mout_i;
wire   [`W-1:0]  W_r, W_i; 
wire   [1:0]     W_ex_r, W_ex_i;
wire			 scaling_bit = 1;
assign Min_r = (k[4] == 0) ? Fout_r : A_r;
assign Min_i = (k[4] == 0) ? Fout_i : A_i;
assign Fin_r = (k[4] == 0) ? X_r : B_r;
assign Fin_i = (k[4] == 0) ? X_i : B_i;
assign Mbf_r = (k[4] == 1) ? A_r : Mout_r;
assign Mbf_i = (k[4] == 1) ? A_i : Mout_i;
FIFO_ST2 FIFO_ST2(CLK, Fin_r, Fin_i, Fout_r, Fout_i);
TWIDDLE_ST2 TWIDDLE_ST2(k[3:0], W_ex_r, W_r, W_ex_i, W_i);
CADD CADD(Fout_r, Fout_i, X_r, X_i, A_r, A_i, B_r, B_i, scaling_bit);
CMUL CMUL(Min_r, Min_i, W_r, W_ex_r, W_i, W_ex_i, Mout_r, Mout_i);
always@(posedge CLK)
begin
    Y_r <= Mbf_r;
    Y_i <= Mbf_i;
end
endmodule


module BUTTERFLY_STAGE3(CLK, k, X_r, X_i, Y_r, Y_i);
input  CLK;
input  [`N_AD_BUS-1:0]  k;
input  [`W-1:0]  X_r, X_i;
output [`W-1:0]  Y_r, Y_i;
reg    [`W-1:0]  Y_r, Y_i;
wire   [`W-1:0]  A_r, A_i, B_r, B_i;
wire   [`W-1:0]  Fin_r, Fin_i;
wire   [`W-1:0]  Fout_r, Fout_i;
wire   [`W-1:0]  Min_r,  Min_i;
wire   [`W-1:0]  Mbf_r, Mbf_i;
wire   [`W-1:0]  Mout_r, Mout_i;
wire   [`W-1:0]  W_r, W_i; 
wire   [1:0]     W_ex_r, W_ex_i;
wire			 scaling_bit = 1;
assign Min_r = (k[3] == 0) ? Fout_r : A_r;
assign Min_i = (k[3] == 0) ? Fout_i : A_i;
assign Fin_r = (k[3] == 0) ? X_r : B_r;
assign Fin_i = (k[3] == 0) ? X_i : B_i;
assign Mbf_r = (k[3] == 1) ? A_r : Mout_r;
assign Mbf_i = (k[3] == 1) ? A_i : Mout_i;
FIFO_ST3 FIFO_ST3(CLK, Fin_r, Fin_i, Fout_r, Fout_i);
TWIDDLE_ST3 TWIDDLE_ST3(k[2:0], W_ex_r, W_r, W_ex_i, W_i);
CADD CADD(Fout_r, Fout_i, X_r, X_i, A_r, A_i, B_r, B_i, scaling_bit);
CMUL CMUL(Min_r, Min_i, W_r, W_ex_r, W_i, W_ex_i, Mout_r, Mout_i);
always@(posedge CLK)
begin
    Y_r <= Mbf_r;
    Y_i <= Mbf_i;
end
endmodule


module BUTTERFLY_STAGE4(CLK, k, X_r, X_i, Y_r, Y_i);
input  CLK;
input  [`N_AD_BUS-1:0]  k;
input  [`W-1:0]  X_r, X_i;
output [`W-1:0]  Y_r, Y_i;
reg    [`W-1:0]  Y_r, Y_i;
wire   [`W-1:0]  A_r, A_i, B_r, B_i;
wire   [`W-1:0]  Fin_r, Fin_i;
wire   [`W-1:0]  Fout_r, Fout_i;
wire   [`W-1:0]  Min_r,  Min_i;
wire   [`W-1:0]  Mbf_r, Mbf_i;
wire   [`W-1:0]  Mout_r, Mout_i;
wire   [`W-1:0]  W_r, W_i; 
wire   [1:0]     W_ex_r, W_ex_i;
wire			 scaling_bit = 1;
assign Min_r = (k[2] == 0) ? Fout_r : A_r;
assign Min_i = (k[2] == 0) ? Fout_i : A_i;
assign Fin_r = (k[2] == 0) ? X_r : B_r;
assign Fin_i = (k[2] == 0) ? X_i : B_i;
assign Mbf_r = (k[2] == 1) ? A_r : Mout_r;
assign Mbf_i = (k[2] == 1) ? A_i : Mout_i;
FIFO_ST4 FIFO_ST4(CLK, Fin_r, Fin_i, Fout_r, Fout_i);
TWIDDLE_ST4 TWIDDLE_ST4(k[1:0], W_ex_r, W_r, W_ex_i, W_i);
CADD CADD(Fout_r, Fout_i, X_r, X_i, A_r, A_i, B_r, B_i, scaling_bit);
CMUL CMUL(Min_r, Min_i, W_r, W_ex_r, W_i, W_ex_i, Mout_r, Mout_i);
always@(posedge CLK)
begin
    Y_r <= Mbf_r;
    Y_i <= Mbf_i;
end
endmodule


module BUTTERFLY_STAGE5(CLK, k, X_r, X_i, Y_r, Y_i);
input  CLK;
input  [`N_AD_BUS-1:0]  k;
input  [`W-1:0]  X_r, X_i;
output [`W-1:0]  Y_r, Y_i;
reg    [`W-1:0]  Y_r, Y_i;
wire   [`W-1:0]  A_r, A_i, B_r, B_i;
wire   [`W-1:0]  Fin_r, Fin_i;
wire   [`W-1:0]  Fout_r, Fout_i;
wire   [`W-1:0]  Min_r,  Min_i;
wire   [`W-1:0]  Mbf_r, Mbf_i;
wire   [`W-1:0]  Mout_r, Mout_i;
wire   [`W-1:0]  W_r, W_i; 
wire   [1:0]     W_ex_r, W_ex_i;
wire			 scaling_bit = 1;
assign Min_r = (k[1] == 0) ? Fout_r : A_r;
assign Min_i = (k[1] == 0) ? Fout_i : A_i;
assign Fin_r = (k[1] == 0) ? X_r : B_r;
assign Fin_i = (k[1] == 0) ? X_i : B_i;
assign Mbf_r = (k[1] == 1) ? A_r : Mout_r;
assign Mbf_i = (k[1] == 1) ? A_i : Mout_i;
FIFO_ST5 FIFO_ST5(CLK, Fin_r, Fin_i, Fout_r, Fout_i);
TWIDDLE_ST5 TWIDDLE_ST5(k[0], W_ex_r, W_r, W_ex_i, W_i);
CADD CADD(Fout_r, Fout_i, X_r, X_i, A_r, A_i, B_r, B_i, scaling_bit);
CMUL CMUL(Min_r, Min_i, W_r, W_ex_r, W_i, W_ex_i, Mout_r, Mout_i);
always@(posedge CLK)
begin
    Y_r <= Mbf_r;
    Y_i <= Mbf_i;
end
endmodule


module BUTTERFLY_STAGE6(CLK, k, X_r, X_i, Y_r, Y_i);
input  CLK;
input  [`N_AD_BUS-1:0]  k;
input  [`W-1:0]  X_r, X_i;
output [`W-1:0]  Y_r, Y_i;
reg    [`W-1:0]  Y_r, Y_i;
wire   [`W-1:0]  A_r, A_i, B_r, B_i;
wire   [`W-1:0]  Fin_r, Fin_i;
wire   [`W-1:0]  Fout_r, Fout_i;
wire   [`W-1:0]  Min_r,  Min_i;
wire   [`W-1:0]  Mbf_r, Mbf_i;
wire			 scaling_bit = 0;
assign Min_r = (k[0] == 0) ? Fout_r : A_r;
assign Min_i = (k[0] == 0) ? Fout_i : A_i;
assign Fin_r = (k[0] == 0) ? X_r : B_r;
assign Fin_i = (k[0] == 0) ? X_i : B_i;
assign Mbf_r = Min_r;
assign Mbf_i = Min_i;
FIFO_ST6 FIFO_ST6(CLK, Fin_r, Fin_i, Fout_r, Fout_i);
CADD CADD(Fout_r, Fout_i, X_r, X_i, A_r, A_i, B_r, B_i, scaling_bit);
always@(posedge CLK)
begin
    Y_r <= Mbf_r;
    Y_i <= Mbf_i;
end
endmodule

module CADD(X_r, X_i, Y_r, Y_i, A_r, A_i, B_r, B_i, scaling_bit);
input  signed [`W-1:0]  X_r, X_i, Y_r, Y_i;
output signed [`W-1:0]  A_r, A_i, B_r, B_i;
input  scaling_bit;
wire   signed [`W:0]  m_r, m_i, n_r, n_i;
assign m_r = X_r + Y_r;
assign m_i = X_i + Y_i;
assign n_r = X_r - Y_r;
assign n_i = X_i - Y_i;
assign A_r = (scaling_bit == 1) ? m_r >>> 1 : m_r;
assign A_i = (scaling_bit == 1) ? m_i >>> 1 : m_i;
assign B_r = (scaling_bit == 1) ? n_r >>> 1 : n_r;
assign B_i = (scaling_bit == 1) ? n_i >>> 1 : n_i;
endmodule
module CMUL(X_r, X_i, W_r, W_ex_r, W_i, W_ex_i, Y_r, Y_i);
input  [`W-1:0]  X_r, X_i;
input  [`W-1:0]  W_r, W_i;
input  [1:0]   W_ex_r, W_ex_i;
output [`W-1:0]  Y_r, Y_i;
wire   [2*`W-2:0]  m_rr, m_ri, m_ir, m_ii;
wire   signed [2*`W-2:0]  n_rr, n_ri, n_ir, n_ii;
wire   [2*`W-1:0]  o_r, o_i;
wire   signed [2*`W-2:0]  a_r, a_i;
wire 	zero = 0;
assign m_rr = FIX_MUL(W_r, X_r);
assign m_ri = FIX_MUL(W_i, X_r);
assign m_ir = FIX_MUL(W_r, X_i);
assign m_ii = FIX_MUL(W_i, X_i);
assign a_r  = {X_r, {`W-1{zero}}};
assign a_i  = {X_i, {`W-1{zero}}};
assign n_rr = (W_ex_r == 1) ?   a_r :
              (W_ex_r == 2) ?  -a_r :
                                m_rr;
assign n_ri = (W_ex_i == 1) ?   a_r :
              (W_ex_i == 2) ?  -a_r :
                                m_ri;
assign n_ir = (W_ex_r == 1) ?   a_i :
              (W_ex_r == 2) ?  -a_i :
                                m_ir;
assign n_ii = (W_ex_i == 1) ?   a_i :
              (W_ex_i == 2) ?  -a_i :
                                m_ii;
assign o_r = n_rr - n_ii;
assign o_i = n_ir + n_ri;
assign Y_r = o_r[2*`W-2:`W-1];
assign Y_i = o_i[2*`W-2:`W-1];
function [2*`W-2:0] FIX_MUL;
	input signed [`W-1:0] in1;
	input signed [`W-1:0] in2;
	reg signed [2*`W-2:0] qs;
	begin
   	qs  = in1 * in2;
   	FIX_MUL = qs;
end
endfunction
endmodule
